import sys

from src.classes import Options, Session
from src.helpers.downloads import download
from src.helpers.listings import get_listing, list_all_projects


def main():
    options = Options(sys.argv)
    session = Session(options)
    if options.get_projects:
        list_all_projects(session)
    if options.listing:
        get_listing(session)
    elif options.download or options.download_all:
        download(session)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print()
        exit(1)



from .classes import MyCookie, options, session
from .helpers import (
    downloads,
    eta_readable,
    listings,
    parallel_download,
    print_functions,
    sizeofmetric,
    url,
    usage,
)



from typing import Union


def size_of_metric_fmt(num: Union[int, float], suffix: str = "B") -> str:
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1000.0
    return "%.1f %s%s" % (num, "Y", suffix)



import time
from multiprocessing import Process
from typing import List

from src.helpers.eta_readable import human_readable_eta
from src.helpers.print_functions import print_info
from src.helpers.sizeofmetric import size_of_metric_fmt


def download_parallel(session, dl_list: List[list], dl_sum: int) -> None:
    """
        Multithreading of download. It starts by making a process of every download by setting the target to
        session.download_file with the parameters of the file. It will then start the processes until the capacity is reached,
        or there are no more left over processes. It will then add the size of downloaded chunks to the total amount downloaded.
        And close no longer active processes. It will then print the progress.
    :param session: The session object.
    :param dl_list: The list of lists representing a file each. Index 0 is the url, 1 is the size in bytes, 2 is the
                    full filepath to give the downloaded file a name and place it in the correct directory.
    :param dl_sum: The total amount of bytes of all files to be downloaded.
    :return: None
    """

    current_processes_amount = 0  # The amount of processes currently active.
    max_processes = int(
        session.options.threads
    )  # The maximum amount of processes that are allowed to be active at a time.
    number_of_processes = len(dl_list)  # The amount of processes.
    finished_processes_amount = 0  # The amount of finished processes.
    current_process_index = (
        0  # The index of the current process in the list of processes.
    )
    downloaded_bytes = 0  # The total amount of downloaded bytes.
    processes = []

    for dl in dl_list:
        processes.append(Process(target=session.download_file, args=dl))

    start = time.time()
    started = []
    while True:
        if (
            current_processes_amount < max_processes
            and finished_processes_amount < number_of_processes
            and current_process_index < number_of_processes
        ):
            processes[current_process_index].start()
            started.append(processes[current_process_index])
            current_process_index += 1
            current_processes_amount += 1
        if (
            current_processes_amount < max_processes
            and current_process_index < number_of_processes
        ):
            continue

        if finished_processes_amount == number_of_processes:
            print_info("\nDownloading complete")
            break

        status = session.queue.get()
        downloaded_bytes += status[0]
        for process in started:
            if not process.is_alive():
                if process.exitcode is not None:
                    process.close()
                    started.remove(process)

        if status[1]:
            current_processes_amount -= 1
            finished_processes_amount += 1
        rate = downloaded_bytes // (time.time() - start)
        if dl_sum > 100:  # Preventing division by zero errors.
            estimated_time_of_arrival = "Never"  # It was this or "After the heat death of the universe" by mmterpstra.
            if rate > 0:
                estimated_time_of_arrival = human_readable_eta(
                    (dl_sum - downloaded_bytes) / rate
                )
            print(
                "\r",
                str(round(downloaded_bytes / dl_sum * 100)) + "%",
                "Downloading",
                size_of_metric_fmt(downloaded_bytes),
                "of",
                size_of_metric_fmt(dl_sum),
                str(size_of_metric_fmt(rate)) + "/sec",
                "ETA:",
                estimated_time_of_arrival,
                end="     ",
            )

        if finished_processes_amount == number_of_processes:
            print_info("\nDownloading complete")
            break



import sys

if sys.version_info >= (3, 10, 0):
    from terminalcolorpy import colored, printcolor


def print_error(text: str) -> None:
    if sys.version_info >= (3, 10, 0):
        printcolor({"text": text, "color": "red"})
    else:
        print(text)


def print_info(text: str) -> None:
    if sys.version_info >= (3, 10, 0):
        printcolor({"text": text, "color": "cyan"})
    else:
        print(text)


def print_warning(text: str) -> None:
    if sys.version_info >= (3, 10, 0):
        printcolor({"text": text, "color": "yellow"})
    else:
        print(text)


def print_file(text: str) -> None:
    if sys.version_info >= (3, 10, 0):
        printcolor({"text": text, "color": "green"})
    else:
        print(text)


def print_rec(dic, depth: int) -> None:
    """
        Prints the folder structure as returned from the api.
    :param dic: An iterable containing dictionaries with the keys "children", "size" and "name".
    :param depth: The recursive depth.
    :return: None
    """
    for item in dic:
        if item["type"] == "directory":
            for i in range(depth * 2):
                print("  ", end="")
            if sys.version_info >= (3, 10, 0):
                print("└──", colored(text=item["name"], color="cyan"))
            else:
                print("└── " + item["name"])
            print_rec(item["children"], depth + 1)
        else:
            for i in range(depth * 2):
                print("  ", end="")

            if sys.version_info >= (3, 10, 0):
                print(
                    "├──",
                    colored(text=item["name"], color="yellow"),
                    "Size: ",
                    colored(text=str(item["size"]), color="red"),
                )
            else:
                print(
                    "├── " + item["name"] + " Size:  " + str(item["size"]),
                )



GSPORT_VERSION = "2.0.1"
ALL_PROJECTS_API = "get_projects/"
LOGIN_URL = "/login/"
LOGGED_IN_URL = "/logged_in_api/"
TWO_FACTOR_AUTH_URL = "/otp_ok/"
DOWNLOAD_RECURSIVE = "/data_api_recursive/"
LOGOUT_URL = "/accounts/logout/"
HOST_URL = "https://portal.genomescan.nl/"
VERIFY_FILES_URL = "/gen_session_file/"
DOWNLOAD_FILE_URL = "/data_api2/"

LIST_EXAMPLE_MESSAGE = """
example usage:
  gsport list 100000                     shows all the files associated with the project from the top level but not any directory or files within directories
  gsport list 100000 -m                  shows all the folders/directories under that project but not their content
  gsport list 100000 -d directory        shows all the files under a directory for that project, no subfolder or files in there
  gsport list 100000 -m -d directory     shows all the files and folders under a directory for that project but not the content of the folders
  gsport list 100000 -r                  shows all the files and folders in a tree like structure
  gsport list 100000 -r -d directory     shows all the files and folders in a tree like structure under a directory for that project
"""
DOWNLOAD_ALL_EXAMPLE_MESSAGE = """
example usage:
  gsport all 100000                    downloads al the files associated with the project from the top level but not any files within directories
  gsport all 100000 -d directory       downloads only the files directly under a directory, no subfolder or files in there
  gsport all 100000 -r                 downloads all the files and folders for that project
  gsport all 100000 -r -d directory    downloads all the files and folders under a directory for that project
  gsport all 100000 -o outputdir       downloads all the files associated with the project from the top level but not any files within directories and outputs them in a directory called outputdir
"""
DOWNLOAD_EXAMPLE_MESSAGE = """
example usage:
  gsport download 100000 test.txt test2.txt                     downloads the specific files, they have to be in the root directory
  gsport download 100000 test.txt -o outputdir                  downloads the specific files and outputs them in a directory called outputdir
"""



import os
from typing import Dict, List, Union

import requests

from src.helpers.listings import get_list
from src.helpers.print_functions import print_error, print_file, print_warning
from src.helpers.url import get_url
from src.variables import DOWNLOAD_FILE_URL, DOWNLOAD_RECURSIVE


def _get_file_names(files: List[Dict[str, str]]) -> Dict[str, Dict[str, str]]:
    file_names = {}
    for file in files:
        file_names[file["name"]] = file
    return file_names


def download(session) -> None:
    """
        Verify the files specified with the download command and passes the approved files to the get url.
    :param session: The session object.
    :return: None
    """
    datafiles = []
    if session.options.download:
        response = requests.get(
            session.options.host + DOWNLOAD_FILE_URL + session.options.project + "/n",
            cookies=session.cookies,
            params={"cd": session.options.dir},
        )
        datafiles = response.json()
    elif session.options.download_all and session.options.recursive:
        response = requests.get(
            session.options.host + DOWNLOAD_RECURSIVE + session.options.project,
            cookies=session.cookies,
            params={"cd": session.options.dir},
        )
        datafiles = get_list(response.text, session.options.dir)
    elif session.options.download_all:
        response = requests.get(
            session.options.host + DOWNLOAD_FILE_URL + session.options.project + "/n",
            cookies=session.cookies,
            params={"cd": session.options.dir},
        )
        datafiles = response.json()
    else:
        exit(1)

    if response.status_code != 200:
        print_error(response.text)
        exit(1)

    if session.options.download:
        requested = session.options.download
        allowed = _get_file_names(datafiles)

        datafiles = []
        for file in requested:
            if file not in allowed:
                print_warning(
                    f"WARNING: {file} is not a valid file for download, make sure the path is spelled correctly."
                )
                continue
            if file in allowed:
                print_file(file)
                datafiles.append(allowed[file])
        if not len(datafiles) > 0:
            print("No valid files to download")
            exit(1)
        if input("Continuing on with the download of the existing files? (y/n)") != "y":
            exit(1)

    if not os.path.isdir(
        session.options.output
    ):  # Create the output folder if it doesn't exist.
        os.makedirs(session.options.output)
    if session.options.download_all and session.options.recursive:
        make_directories(datafiles, output=session.options.output)
    get_url(session, datafiles)


def make_directories(
    files: List[Dict[str, Union[str, int]]],
    directory_path_length: int = 0,
    output: str = ".",
) -> None:
    """
        Create the directories that the files will be put in.
    :param files: The list of dictionaries containing file information.
    :param directory_path_length:
    :param output: The output directory.
    :return: None
    """
    for file in files:  # Go through every file.
        total_path = output  # Begin the path with the output directory.
        for path in file["name"].split("/")[
            directory_path_length:-1
        ]:  # Loop through the elements of the path, but not the file. This works because the server should never return "\" based paths.
            total_path = os.path.join(
                total_path, path
            )  # Append the path element to the total path.
            if not os.path.isdir(
                total_path
            ):  # Create the directory with the using the total path if it doesn't exist yet.
                try:
                    os.makedirs(total_path)
                except FileExistsError:
                    pass  # This can be the case with multithreading.



import unittest
from unittest.mock import patch

import requests_mock

from main import main
from src.classes import Session

"""
Tests for the login process

Author: Gonzalo Vela
"""


class TestLogin(unittest.TestCase):
    @patch("sys.argv", ["script_name"])
    @patch.object(Session, "login")
    def test_login_false(self, mock_login):
        """
        Test that unauthorized cookies trigger the login process.
        """

        with requests_mock.Mocker() as m:
            m.get(
                "https://portal.genomescan.nl//logged_in_api/",
                text='{"logged_in": false}',
                status_code=200,
            )

            main()

            mock_login.assert_called_once()



import json
import os
import time
from getpass import getpass
from multiprocessing import Queue
from pathlib import Path

import requests

from src.classes import MyCookieJar
from src.helpers.eta_readable import human_readable_eta
from src.helpers.print_functions import print_error, print_info, print_warning
from src.helpers.sizeofmetric import size_of_metric_fmt
from src.variables import (
    GSPORT_VERSION,
    HOST_URL,
    LOGGED_IN_URL,
    LOGIN_URL,
    LOGOUT_URL,
    TWO_FACTOR_AUTH_URL,
)


class Session:
    def __init__(self, options):
        self.options = options
        self.cookies = MyCookieJar(
            filename=os.path.join(str(Path.home()), ".gs_cookies.txt")
        )
        self.queue = Queue()
        self.process = Queue()

        if options.clear_cookies:
            self.logout()
        try:
            self.cookies.load()
            response = requests.get(
                options.host + LOGGED_IN_URL, cookies=self.cookies
            ).text
            if json.loads(response)["logged_in"]:
                print_info("[session] cookies found.")
            else:
                self.login()
        except FileNotFoundError:
            print_info("[session] No cookies found. Logging in...")
            self.login()

    def login(self) -> None:
        """
            This function walks the user through the login procedure, starting by making a session, followed by getting
            the csrf cookie. The user is then prompted to put in a username/email and password. They are then verified
            by the server. If successful, the server sends een login-token trough email to the user, which gsport then
            requests. After a successful authorisation, the cookie is saved.
        :return:
        """
        print_info("[login] Opening session...")
        session = requests.Session()  # Make a session.
        session.cookies = MyCookieJar(
            os.path.join(str(Path.home()), ".gs_cookies.txt")
        )  # set the cookie.
        print_info("[login] Get login page")
        # Perform a GET request to obtain the CSRF token
        response = session.get(HOST_URL + LOGIN_URL)
        csrftoken = response.cookies["csrftoken"]
        success = False
        while not success:
            username = input("Username: ")
            psw = getpass()
            login_data = dict(
                username=username, password=psw, csrfmiddlewaretoken=csrftoken, next="/"
            )
            response = session.post(
                self.options.host + LOGIN_URL,
                data=login_data,
                headers=dict(Referer=self.options.host + LOGIN_URL),
            )
            # try to log in.
            if response.status_code != 200:
                print_warning(response.text)
                continue
            login_data = dict(
                token=input("Token: "),
                password=psw,
                username=username,
                csrfmiddlewaretoken=csrftoken,
                next="/",
            )
            response = session.post(
                self.options.host + TWO_FACTOR_AUTH_URL,
                data=login_data,
                headers={
                    "Referer": self.options.host + LOGIN_URL,
                    "User-Agent": "gsport " + GSPORT_VERSION,
                },
            )
            if response.status_code != 200:
                print_error(response.text)
                continue
            success = True

        print_info("[login] Success, saving cookies...")
        session.cookies.save(ignore_discard=True)

        print_info("[login] Done.")
        self.cookies = session.cookies

    def download_file(self, url: str, fsize: int, fname: str) -> None:
        """
            Download the file by streaming the dat from the url.
        :param url: The download link.
        :param fsize: The file size in bytes.
        :param fname: The filename.
        :return: None
        """
        try:
            dsize = 0
            start = time.time()

            with requests.get(
                url, stream=True, cookies=self.cookies
            ) as r:  # Start the download.
                self.options.dir = "/".join(self.options.dir.split("/")[:-1])

                if self.options.dir == "":
                    self.options.dir = ""
                with open(fname, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)

                            dsize += len(chunk)
                            rate = dsize // (time.time() - start)
                            if not self.options.download_all:
                                print(
                                    "\r"
                                    + size_of_metric_fmt(fsize)
                                    + " "
                                    + str(round(dsize / fsize * 100))
                                    + "% "
                                    + str(size_of_metric_fmt(rate))
                                    + "/sec ",
                                    "ETA:",
                                    human_readable_eta((fsize - dsize) / rate),
                                    end="     ",
                                )
                            else:
                                self.queue.put([len(chunk), False])

            self.queue.put([0, True])
        except KeyboardInterrupt:
            return
        return

    def logout(self) -> None:
        try:
            self.cookies.load()
            response = requests.get(
                self.options.host + LOGOUT_URL, cookies=self.cookies
            )
            if response.status_code == 200:
                print_info("[logout] Logged out.")
            else:
                print_error("[logout] Error logging out.")
                exit(1)
            # TODO: Add code to delete cookie from system.
        except FileNotFoundError:
            print_info("[session] No cookies found to clear. exiting...")
        exit(0)



import argparse
import os

from src.helpers import print_functions
from src.variables import (
    DOWNLOAD_ALL_EXAMPLE_MESSAGE,
    DOWNLOAD_EXAMPLE_MESSAGE,
    GSPORT_VERSION,
    HOST_URL,
    LIST_EXAMPLE_MESSAGE,
)


class Options:
    def __init__(self, argv):
        # Create the parser for gsport.
        parser = argparse.ArgumentParser(
            prog="gsport",
            description="GSPORT command-line tool for accessing GenomeScans new Customer Portal",
        )
        subparsers = parser.add_subparsers(
            title="subcommands",
            description="valid subcommands",
            help="use subcommand with -h for additional help",
            dest="subparser_name",
        )

        # Parser commands.
        parser.add_argument(
            "-H",
            "--host",
            default=HOST_URL,
            help="The host site default is %(default)s",
        )
        parser.add_argument(
            "-c",
            "--clear-cookies",
            action="store_true",
            help="clear cookies and logout session",
        )
        parser.add_argument(
            "-v",
            "--version",
            help="show the software version and exit",
            action="version",
            version=f"%(prog)s {GSPORT_VERSION}",
        )
        parser.add_argument(
            "-p",
            "--projects",
            help="show all the projects that a user has access to",
            action="store_true",
        )

        # List of shared commands for subcommands.
        project_parser = argparse.ArgumentParser(add_help=False)
        project_parser.add_argument(
            "PROJECT", help="[projectcode] for specific projects"
        )
        download_and_listing_shared = argparse.ArgumentParser(add_help=False)
        download_and_listing_shared.add_argument(
            "-d",
            "--cd",
            default=".",
            help="files (or directories) in dir, dirs can be appended with forward"
            " slashes: / (eg. 'Analysis/Sample 1', with quotes) or Analysis/s1/bam"
            " (without spaces, no quotes needed)",
        )
        downloads_shared = argparse.ArgumentParser(add_help=False)
        downloads_shared.add_argument(
            "-o",
            "--output",
            default=".",
            help="directory that downloaded files will go in, default is current directory",
        )

        # List subcommand.
        subparser_list = subparsers.add_parser(
            "list",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="prints the output",
            description="this subcommand prints the output",
            epilog=LIST_EXAMPLE_MESSAGE,
            parents=[project_parser, download_and_listing_shared],
        )
        list_options_group = subparser_list.add_mutually_exclusive_group()
        list_options_group.add_argument(
            "-m", "--dirs", action="store_true", help="show directories"
        )
        list_options_group.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            help="recursive complete tree from --cd "
            "[dir] or everything if no --cd option is given ",
        )
        subparser_list.set_defaults(func=self.L)

        # Download subcommand.
        subparser_download = subparsers.add_parser(
            "download",
            help="download specified files",
            description="this allows the download of individual files, use the full path for files",
            epilog=DOWNLOAD_EXAMPLE_MESSAGE,
            parents=[project_parser, downloads_shared],
        )
        subparser_download.add_argument(
            "FILE", help="files to download seperated by spaces", nargs="+"
        )
        subparser_download.set_defaults(func=self.D)

        # Download all subcommand.
        subparser_download_all = subparsers.add_parser(
            "all",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help="download all files",
            description="this subcommand allows the download of all files",
            epilog=DOWNLOAD_ALL_EXAMPLE_MESSAGE,
            parents=[project_parser, download_and_listing_shared, downloads_shared],
        )
        subparser_download_all.add_argument(
            "-t", "--threads", type=int, default=os.cpu_count()
        )
        subparser_download_all.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            help="recursive complete tree from --cd "
            "[dir] or everything if no --cd option is given ",
        )
        subparser_download_all.set_defaults(func=self.A)

        # Parse arguments.
        args = parser.parse_args()

        self.host: str = (
            args.host
        )  # The host site that gsport should make connection to.
        self.download: list | None = None  # When the download option is being used the files are saved in a list.
        self.download_all: bool = False  # Is the all option being used.
        self.listing: bool = False  # Is the list option being used.
        self.recursive: bool = False  # Is the recursive option being used.
        self.project: str | None = (
            None  # The projectcode that the user requests things from.
        )
        self.clear_cookies: bool = (
            args.clear_cookies
        )  # Clear the cookies, must be logged in to clear cookies.
        self.threads: int = (
            1  # The amount of threads being used for multithreading, Linux only.
        )
        self.folder_mode: bool = False  # Show only directories in list sub-command.
        self.dir: str = ""
        self.get_projects: bool = args.projects  # Show all user projects?
        self.output: str = "."  # The directory that the files are being saved to.

        if args.host != parser.get_default("host"):
            print_functions.print_info(f"Using alternative host {args.host}")

        if args.subparser_name is not None:
            args.func(args)

    def L(self, args):
        """
            Gets called when the subcommand L is used.
        :param args: The argument Namespace object.
        :return: None
        """
        self.folder_mode = args.dirs
        self.dir = args.cd + "/"
        self.recursive = args.recursive
        self.project = args.PROJECT
        self.listing = True

    def A(self, args):
        """
            Gets called when the subcommand A is used.
        :param args: The argument Namespace object.
        :return: None
        """
        self.dir = args.cd + "/"
        self.recursive = args.recursive
        self.threads = args.threads
        self.project = args.PROJECT
        self.download_all = True
        self.output = args.output

    def D(self, args):
        """
            Gets called when the subcommand D is used.
        :param args: The argument Namespace object.
        :return: None
        """
        self.download = args.FILE
        self.project = args.PROJECT
        self.threads = os.cpu_count()
        self.output = args.output



def human_readable_eta(seconds):
    days = seconds // 86400
    hours = seconds // 3600 % 24
    minutes = seconds // 60 % 60
    seconds = seconds % 60
    ret = str(round(days)) + "d" if days > 0 else ""
    ret += str(round(hours)) + "h" if hours > 0 else ""
    ret += str(round(minutes)) + "m" if minutes > 0 else ""
    ret += str(round(seconds)) + "s" if seconds > 0 and minutes < 1 else ""
    return ret



import threading
from http.cookiejar import MozillaCookieJar

"""
Custom CookieJar to fix multiprocessing issues on Windows
"""


class MyCookieJar(MozillaCookieJar):
    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_cookies_lock"]
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._cookies_lock = threading.RLock()



import json
import os
import sys

import requests

if sys.version_info >= (3, 10, 0):
    from terminalcolorpy import colored

from src.helpers.print_functions import print_error, print_info, print_rec
from src.variables import ALL_PROJECTS_API, DOWNLOAD_RECURSIVE


def get_listing(session) -> None:
    """
        Gets the json in the form {data: [{name:str, size: str, children: list[dict]}]}
        and prints the values in a certain way depending on it being default, -m or -r.
    :param session: Session object.
    :return: None
    """
    response = requests.get(
        session.options.host + DOWNLOAD_RECURSIVE + session.options.project,
        cookies=session.cookies,
        params={"cd": session.options.dir},
    )

    try:
        datafiles = json.loads(response.text)
    except json.decoder.JSONDecodeError:
        print_error(f"[get_listing] Error reading response: {response.text}")
        exit(1)

    if session.options.recursive:
        print_rec(datafiles["children"], 0)
    else:
        if not session.options.folder_mode:
            for file in datafiles["children"]:
                if file["type"] == "file":
                    if sys.version_info >= (3, 10, 0):
                        print(
                            colored(text=file["name"], color="yellow"),
                            "Size: ",
                            colored(text=str(file["size"]), color="red"),
                        )
                    else:
                        print(file["name"] + " Size:  " + str(file["size"]))
        else:
            for file in datafiles["children"]:
                if len(file["name"]) > 0:
                    if file["type"] == "directory":
                        if sys.version_info >= (3, 10, 0):
                            print(colored(text=file["name"], color="cyan"))
                        else:
                            print(file["name"])
                    else:
                        if sys.version_info >= (3, 10, 0):
                            print(colored(text=file["name"], color="yellow"))
                        else:
                            print(file["name"])


def list_all_projects(session) -> None:
    """
        Prints all the projects a user has access to.
    :param session: Session object.
    :return:
    """
    print_info("[requesting projects]")
    response = requests.get(
        session.options.host + ALL_PROJECTS_API, cookies=session.cookies
    )
    try:
        projects = response.json()
        for i in projects["projects"]:
            print(i)
        if response.status_code != 200:
            exit(1)
    except (json.decoder.JSONDecodeError, KeyError):
        print_error(f"[get_listing] Error reading response: {response.text}")
        exit(1)


def get_list(res, session_dir):
    flist = []

    def print_list(dic, path):
        for item in dic:
            if item["type"] == "directory":
                d = os.path.join(path, item["name"])
                if not os.path.isdir(d):
                    try:
                        os.makedirs(d)
                    except FileExistsError:
                        pass  # this can be the case with multithreading
                print_list(item["children"], d)
            else:
                flist.append({"name": path + "/" + item["name"], "size": item["size"]})

    print_list(json.loads(res)["children"], session_dir)
    return flist



def usage():
    print(
        """
Usage: gsport [options]
GSPORT command-line tool for accessing GenomeScan Customer Portal

Options
-H --host [host], default: https://portal.genomescan.nl
-p --project [project] project (required with -l, -d, -a)
-l --list return a list of all files
-s --size return a list with the size
-d --download [filename]
-a --download-all download all files from project -p or --project
-f --force downloading files even if they already exist
-c --clear-cookies clear session/cookies
-t --workers [n] allow n concurrent workers (defaults to number of logical cpu cores) (works only on Linux)
--dirs show directories instead of files (combined with -l or --list)
--cd [dir] show files (or directories) in dir,
     dirs can be appended with forward slashes: / (eg. "Analysis/Sample 1", with quotes)
     or Analysis/s1/bam (without spaces, no quotes needed)
-r --recursive lists/downloads complete tree from --cd [dir] or everything if no --cd option is given
-h --help prints this help
-v --version show gsport version
-i --ignore Ignore MD5 checksum result and download only files not on the system
-I --includeFile Path to file containing list of files to include in the download. One file per line and save the file using the UTF-8 encoding. only works with the -a flag.
-E --excludeFile Path to file containing list of files to exclude from downloading. One file per line and save the file using the UTF-8 encoding. only works with the -a flag.
-C --checksumFile Path to the file containing a lise of all the file checksums. It will store the md5 checksum of all downloaded files. If the checksum of the file to download is not provided, the script will download the file twice to test if the MD5 match.
-P --path Local location to store the files to.

Note: Using --dirs together with -r / --recursive has no effect

Example usage: gsport -p 100000 -l shows all the files under that project
               gsport -p 100000 -ls shows all the files under that project with the filesize
               gsport -p 100000 -l --dirs shows all the folders/directories under that project
               gsport -p 100000 -l --cd Analysis shows all the files under Analysis for that project
               gsport -p 100000 -l -r shows all the files and folders in Analysis in a tree structure
               gsport -p 100000 -l --dirs cd Analysis shows all the folders under Analysis for that project
               gsport -p 100000 -a -r downloads all the files and folders for that project
               gsport -p 100000 -a -r --cd Analysis downloads all the files and folder under Analysis for that project
               gsport -p 100000 -a --cd Analysis downloads only the files directly under Analysis, no subfolder or files in there.
               gsport -p 100000 -a --cd Analysis/s1 downloads only the files directly under Analysis/s1
               gsport -p 100000 -a -I "C:\\project\\include.txt"
               gsport -p 100000 -a -E "C:\\project\\exclude.txt"
               gsport -p 100000 -a -C "C:\\project\\localChecksums.md5"
               gsport -p 100000 -a -P "C:\\project\\exclude.txt"
"""
    )



import sys
import unittest
from unittest.mock import patch

from src.classes import Options

"""
Tests for the argument parser in the Options class.

Each test checks that specific command-line arguments are correctly
parsed and assign the appropriate attributes.

Author: Gonzalo Vela
"""


class TestArgumentParser(unittest.TestCase):
    @patch("sys.argv", ["script_name", "-H", "test_host"])
    def test_host(self):
        """
        Test of host argument
        """
        options = Options(sys.argv)
        assert options.host == "test_host"

    @patch("sys.argv", ["script_name", "-c"])
    def test_clear_cookies(self):
        """
        Test of clear cookies flag
        """
        options = Options(sys.argv)
        assert options.clear_cookies

    @patch("sys.argv", ["script_name", "list", "999", "-m"])
    def test_set_see_directories(self):
        """
        Test of folder mode flag
        """
        options = Options(sys.argv)
        assert options.folder_mode

    @patch("sys.argv", ["script_name", "list", "999", "-d", "test_dir"])
    def test_set_directory(self):
        """
        Test of custom directory argument
        """
        options = Options(sys.argv)
        assert options.dir == "test_dir/"

    @patch("sys.argv", ["script_name", "list", "999", "-r"])
    def test_set_recursive(self):
        """
        Test of recursive flag
        """
        options = Options(sys.argv)
        assert options.recursive

    @patch("sys.argv", ["script_name", "list", "999", "-m", "-d", "test_dir"])
    def test_set_both_folder_files(self):
        """
        Test of folder mode with custom directory
        """
        options = Options(sys.argv)
        assert options.folder_mode
        assert options.dir == "test_dir/"

    @patch("sys.argv", ["script_name", "list", "999", "-r", "-d", "test_dir"])
    def test_set_both_recursive_files(self):
        """
        Test of recursive flag with custom directory
        """
        options = Options(sys.argv)
        assert options.recursive
        assert options.dir == "test_dir/"

    @patch("sys.argv", ["script_name", "all", "999", "-o", "test_dir"])
    def test_set_output(self):
        """
        Test of output directory argument
        """
        options = Options(sys.argv)
        assert options.output == "test_dir"

    @patch("sys.argv", ["script_name", "all", "999", "-t", "4"])
    def test_set_threads(self):
        """
        Test of thread count argument
        """
        options = Options(sys.argv)
        assert options.threads == 4

    @patch("sys.argv", ["script_name", "all", "999", "-r"])
    def test_set_recursive_download(self):
        """
        Test of recursive download flag
        """
        options = Options(sys.argv)
        assert options.recursive



import re
import sys
import unittest
from io import StringIO
from unittest.mock import mock_open, patch

import requests_mock

from main import main

"""
Tests for the list command

Each tests checks that the list command displays the corresponding file and directories

Author: Gonzalo Vela
"""


def get_mock_cookie_jar():
    return """# Netscape HTTP Cookie File
       # https://curl.se/docs/http-cookies.html
       # This file was generated by hand for testing purposes

       portal.genomescan.nl    TRUE   /  FALSE  0  csrftoken  """


class TestListcommand(unittest.TestCase):
    def remove_ansi_escape_sequences(self, text):
        """
        Helper function to remove escape sequences for the ease of string comparison
        """
        ansi_escape = re.compile(r"\x1b\[[0-9;]*[mK]")

        return ansi_escape.sub("", text)

    @patch("sys.argv", ["script_name", "list", "999", "-m"])
    @patch("http.cookiejar.MozillaCookieJar.load")
    def test_folder_mode(self, mock_load):
        """
        Test list command when folder mode is set
        """
        cookie_mock = mock_open(read_data=get_mock_cookie_jar())

        original_open = open

        with patch("builtins.open", cookie_mock):
            with requests_mock.Mocker() as m:
                m.get(
                    "https://portal.genomescan.nl//logged_in_api/",
                    text='{"logged_in": true}',
                    status_code=200,
                )

                with original_open(
                    "tests/assets/recursive_mock.json", "r"
                ) as recursive_mock_file:
                    recursive_mock = recursive_mock_file.read()

                m.get(
                    "https://portal.genomescan.nl//data_api_recursive/999?cd=.%2F",
                    text=recursive_mock,
                )

                captured = StringIO()
                sys.stdout = captured

                main()

                captured = captured.getvalue()

                expected_output = "[session] cookies found.\ntest_map_salah\ntest_999\n"

                self.assertEqual(
                    self.remove_ansi_escape_sequences(captured), expected_output
                )

    @patch("sys.argv", ["script_name", "list", "999", "-d", "test_999"])
    @patch("http.cookiejar.MozillaCookieJar.load")
    def test_show_files_from_project(self, mock_load):
        """
        Test list command shows files under the projects
        """
        cookie_mock = mock_open(read_data=get_mock_cookie_jar())

        original_open = open

        with patch("builtins.open", cookie_mock):
            with requests_mock.Mocker() as m:
                m.get(
                    "https://portal.genomescan.nl//logged_in_api/",
                    text='{"logged_in": true}',
                    status_code=200,
                )

                with original_open(
                    "tests/assets/project_files.json", "r"
                ) as project_files:
                    project_files = project_files.read()

                m.get(
                    "https://portal.genomescan.nl//data_api_recursive/999?cd=test_999%2F",
                    text=project_files,
                )

                captured = StringIO()
                sys.stdout = captured

                main()

                captured = captured.getvalue()

                expected_output = "[session] cookies found.\ntest_10G.txt Size:  10737418240\ntest2_10G.txt Size:  10737418240\n"

                self.assertEqual(
                    self.remove_ansi_escape_sequences(captured), expected_output
                )

    @patch("sys.argv", ["script_name", "list", "999", "-m", "-d", "test_999"])
    @patch("http.cookiejar.MozillaCookieJar.load")
    def test_show_files_folders_from_project(self, mock_load):
        """
        Test list command shows files and folders under a project
        """
        cookie_mock = mock_open(read_data=get_mock_cookie_jar())

        original_open = open

        with patch("builtins.open", cookie_mock):
            with requests_mock.Mocker() as m:
                m.get(
                    "https://portal.genomescan.nl//logged_in_api/",
                    text='{"logged_in": true}',
                    status_code=200,
                )

                with original_open(
                    "tests/assets/project_files.json", "r"
                ) as project_files:
                    project_files = project_files.read()

                m.get(
                    "https://portal.genomescan.nl//data_api_recursive/999?cd=test_999%2F",
                    text=project_files,
                )

                captured = StringIO()
                sys.stdout = captured

                main()

                captured = captured.getvalue()

                expected_output = "[session] cookies found.\ntest_map_salah\ntest_10G.txt\ntest2_10G.txt\n"

                self.assertEqual(
                    self.remove_ansi_escape_sequences(captured), expected_output
                )

    @patch("sys.argv", ["script_name", "list", "999", "-r", "-d", "test_999"])
    @patch("http.cookiejar.MozillaCookieJar.load")
    def test_show_project_recursively(self, mock_load):
        """
        Tests list command shows files recursively
        """
        cookie_mock = mock_open(read_data=get_mock_cookie_jar())

        original_open = open

        with patch("builtins.open", cookie_mock):
            with requests_mock.Mocker() as m:
                m.get(
                    "https://portal.genomescan.nl//logged_in_api/",
                    text='{"logged_in": true}',
                    status_code=200,
                )

                with original_open(
                    "tests/assets/project_files.json", "r"
                ) as project_files:
                    project_files = project_files.read()

                m.get(
                    "https://portal.genomescan.nl//data_api_recursive/999?cd=test_999%2F",
                    text=project_files,
                )

                captured = StringIO()
                sys.stdout = captured

                main()

                captured = captured.getvalue()

                expected_output = "[session] cookies found.\n└── test_map_salah\n    ├── 3660_Color_palette (1).pdf Size:  517854\n├── test_10G.txt Size:  10737418240\n├── test2_10G.txt Size:  10737418240\n"

                self.assertEqual(
                    self.remove_ansi_escape_sequences(captured), expected_output
                )



import os
from typing import Dict, List, Union

import requests

from src.helpers.parallel_download import download_parallel
from src.variables import VERIFY_FILES_URL


def get_url(session, datafiles: List[Dict[str, Union[str, int]]]) -> None:
    """
        Generate the url of every file to be downloaded and puts them in a list. Call the multithreading download function
        if the os is Linux, otherwise call the download function for each file to be downloaded.
    :param session: Session object.
    :param datafiles: The list of dictionaries representing files with a size in bytes and name: as path.
    :return: None
    """
    dl_list = []
    dl_sum = 0

    for file in datafiles:
        fsize = file["size"] if file["size"] != 0 else 1
        fname = (
            os.path.join(
                session.options.output, file["name"].replace("\\", "/").split("/")[-1]
            )
            if not session.options.recursive
            else os.path.join(session.options.output, os.path.normpath(file["name"]))
        )
        filename = (
            "/"
            + (session.options.dir if not session.options.recursive else "")
            + "/"
            + file["name"]
        )
        dl_sum += fsize
        filename = filename.replace("\\", "/")
        response = requests.get(
            session.options.host + VERIFY_FILES_URL,
            cookies=session.cookies,
            params={"project": session.options.project, "filename": filename},
        )
        url = (
            session.options.host
            + "/session_files2/"
            + session.options.project
            + "/"
            + response.text
        )
        dl_list.append([url, fsize, fname])

    download_parallel(session, dl_list, dl_sum)



import unittest

from src.helpers.sizeofmetric import (  # Adjust to match your structure
    size_of_metric_fmt,
)

"""
Tests for the converter to metric size function

Each test checks that the function returns each size in the correct format

Author: Gonzalo Vela
"""


class TestMetricSize(unittest.TestCase):
    def test_bytes(self):
        """
        Test of only bytes
        """
        result = size_of_metric_fmt(999)
        assert result == "999.0 B"

    def test_kilobyte(self):
        """
        Test of one kilobyte
        """
        result = size_of_metric_fmt(1000)
        assert result == "1.0 KB"

    def test_kilobyte_rounding(self):
        """
        Test for kilobyte rounding
        """
        result = size_of_metric_fmt(1536)
        assert result == "1.5 KB"

    def test_megabyte(self):
        """
        Test of one megabyte
        """
        result = size_of_metric_fmt(1_000_000)
        assert result == "1.0 MB"

    def test_gigabyte(self):
        """
        Test of one gigabyte
        """
        result = size_of_metric_fmt(1_000_000_000)
        assert result == "1.0 GB"

    def test_terabyte(self):
        """
        Test of one terabyte
        """
        result = size_of_metric_fmt(1_000_000_000_000)
        assert result == "1.0 TB"

    def test_yottabyte(self):
        """
        Tests if yotabytes are returned
        """
        result = size_of_metric_fmt(1e27)
        assert result == "1000.0 YB"

    def test_float_input(self):
        """
        Test input with decimals
        """
        result = size_of_metric_fmt(123456.78)
        assert result == "123.5 KB"

    def test_custom_suffix(self):
        """
        Test when a custom suffix is given
        """
        result = size_of_metric_fmt(1000, suffix="bps")
        assert result == "1.0 Kbps"

    def test_zero_value(self):
        """
        Test edge case of a zero value
        """
        result = size_of_metric_fmt(0)
        assert result == "0.0 B"

    def test_negative_value(self):
        """
        Test edge case of a negative value
        """
        result = size_of_metric_fmt(-1500)
        assert result == "-1.5 KB"



import unittest

from src.helpers.eta_readable import human_readable_eta

"""
Tests for the converter to human readable time

Each test checks that the function returns each time in the correct format

Author: Gonzalo Vela
"""


class TestETAReadable(unittest.TestCase):
    def test_one_second(self):
        """
        Test of one second
        """
        result = human_readable_eta(1)
        self.assertEqual(result, "1s")

    def test_one_minute(self):
        """
        Test of one minute
        """
        result = human_readable_eta(60)
        self.assertEqual(result, "1m")

    def test_one_hour(self):
        """
        Test of one hour
        """
        result = human_readable_eta(3600)
        self.assertEqual(result, "1h")

    def test_one_day(self):
        """
        Test of one day
        """
        result = human_readable_eta(86400)
        self.assertEqual(result, "1d")

    def test_minute_with_extra_seconds(self):
        """
        Test of one minute with a few seconds
        """
        result = human_readable_eta(65)
        self.assertEqual(result, "1m")

    def test_hour_with_minutes(self):
        """
        Test of one hour with some minutes
        """
        result = human_readable_eta(3900)
        self.assertEqual(result, "1h5m")

    def test_day_and_half(self):
        """
        Test of one day and half a day
        """
        result = human_readable_eta(129600)
        self.assertEqual(result, "1d12h")

    def test_day_half_with_minutes(self):
        """
        Test of one day and half a day with some minutes
        """
        result = human_readable_eta(131400)

        self.assertEqual(result, "1d12h30m")

    def test_zero_time(self):
        """
        Test of zero seconds
        """
        result = human_readable_eta(0)
        self.assertEqual(result, "")
